# Wilfred
# Copyright (C) 2020, Vilhelm Prytz, <vilhelm@prytznet.se>
#
# Licensed under the terms of the MIT license, see LICENSE.
# https://github.com/wilfred-dev/wilfred

import threading


class KeyboardThread(threading.Thread):
    def __init__(self, input_callback, params):
        self.input_callback = input_callback
        self.params = params

        super(KeyboardThread, self).__init__(name="wilfred-console-input", daemon=True)

        self.start()

    def run(self):
        while True:
            self.input_callback(input(), self.params)



# Wilfred
# Copyright (C) 2020, Vilhelm Prytz, <vilhelm@prytznet.se>
#
# Licensed under the terms of the MIT license, see LICENSE.
# https://github.com/wilfred-dev/wilfred

import json
import click

from appdirs import user_config_dir
from os.path import isfile, isdir
from pathlib import Path

from wilfred.message_handler import warning, error

API_VERSION = 1


class Config(object):
    def __init__(self):
        self.data_dir = f"{user_config_dir()}/wilfred"
        self.config_path = f"{self.data_dir}/config.json"

        if not isdir(self.data_dir):
            Path(self.data_dir).mkdir(parents=True, exist_ok=True)

        self._read()

    def _read(self):
        if not isfile(self.config_path):
            warning(
                "configuration file does not exist. Run `wilfred setup` to initiate the program.",
            )
            self.configuration = None

            return

        with open(self.config_path) as f:
            self.configuration = json.loads(f.read())

        if self.configuration["meta"]["version"] != API_VERSION:
            error(
                f"configuration has API level {self.configuration['meta']['version']}, Wilfreds API level is {API_VERSION}",
                exit_code=1,
            )

    def write(self, data_path):
        try:
            Path(data_path).mkdir(parents=True, exist_ok=True)
        except Exception as e:
            error(
                f"unable to access specified path {click.style(str(e), bold=True)}",
                exit_code=1,
            )

        with open(self.config_path, "w") as f:
            f.write(
                json.dumps(
                    {"meta": {"version": API_VERSION}, "data_path": data_path}, indent=4
                )
            )



# Wilfred
# Copyright (C) 2020, Vilhelm Prytz, <vilhelm@prytznet.se>
#
# Licensed under the terms of the MIT license, see LICENSE.
# https://github.com/wilfred-dev/wilfred

from random import choice
from string import ascii_lowercase, digits


def random_string(length=8):
    """
    Generate a random string of fixed length

    :param int length: length of string to generate
    """

    return "".join(choice(ascii_lowercase + digits) for i in range(length))


def is_integer(variable):
    try:
        int(variable)
    except Exception:
        return False

    return True



# Wilfred
# Copyright (C) 2020, Vilhelm Prytz, <vilhelm@prytznet.se>
#
# Licensed under the terms of the MIT license, see LICENSE.
# https://github.com/wilfred-dev/wilfred

import sqlite3
import click

from appdirs import user_data_dir
from os.path import isfile, isdir
from pathlib import Path

from wilfred.message_handler import info, error

API_VERSION = 1


def _query(path, query, fetchone=False):
    """
    executes specified query on sqlite3 database located at specified path

    :param str path: the path of the sqlite3 database
    :param str query: sql query to execute
    """

    result = []

    try:
        conn = sqlite3.connect(path)
        conn.row_factory = sqlite3.Row

        cur = conn.cursor()
        cur.execute(query)

        if fetchone:
            result = cur.fetchone()
        else:
            for r in cur.fetchone() if fetchone else cur.fetchall():
                result.append(dict(r))

        conn.commit()
        conn.close()
    except sqlite3.OperationalError as e:
        error(
            "could not communicate with database " + click.style(str(e), bold=True),
            exit_code=1,
        )
    except sqlite3.IntegrityError as e:
        error("invalid input " + click.style(str(e), bold=True), exit_code=1)

    return result


class Database(object):
    def __init__(self):
        self.data_dir = f"{user_data_dir()}/wilfred"
        self.database_path = f"{self.data_dir}/wilfred.db"

        if not isdir(self.data_dir):
            Path(self.data_dir).mkdir(parents=True, exist_ok=True)

        # main sqlite3 database
        if not isfile(self.database_path):
            info("local database not found, creating")
            self._create_tables()
            self._insert_api_version()

            return
        if (
            int(
                self.query(
                    f"SELECT value FROM constants WHERE name = 'api_version'",
                    fetchone=True,
                )["value"]
            )
            != API_VERSION
        ):
            error(f"database API level differs from Wilfreds", exit_code=1)

    def _create_tables(self):
        _tables = [
            """CREATE TABLE constants (
                name VARCHAR NOT NULL UNIQUE,
                value VARCHAR,
                PRIMARY KEY (name)
            );""",
            """CREATE TABLE servers (
                id VARCHAR NOT NULL UNIQUE,
                name VARCHAR NOT NULL UNIQUE,
                image_uid VARCHAR NOT NULL,
                memory INT NOT NULL,
                port INT NOT NULL UNIQUE,
                custom_startup VARCHAR,
                status VARCHAR NOT NULL,
                PRIMARY KEY (id)
            );""",
            """CREATE TABLE variables (
                server_id VARCHAR NOT NULL,
                variable VARCHAR NOT NULL,
                value VARCHAR NOT NULL
            )""",
        ]

        with click.progressbar(
            _tables, label="Creating tables", length=len(_tables)
        ) as tables:
            for table in tables:
                _query(self.database_path, table)

    def _insert_api_version(self):
        _query(
            self.database_path,
            f"INSERT INTO constants (name, value) VALUES ('api_version', '{API_VERSION}')",
        )

    def query(self, query, *args, **kwargs):
        return _query(self.database_path, query, *args, **kwargs)



# Wilfred
# Copyright (C) 2020, Vilhelm Prytz, <vilhelm@prytznet.se>
#
# Licensed under the terms of the MIT license, see LICENSE.
# https://github.com/wilfred-dev/wilfred


class ContainerVariables(object):
    def __init__(self, server, image, database, install=False):
        self._server = server
        self._image = image
        self._database = database
        self._install = install

    def parse_startup_command(self, cmd):
        for k, v in self.get_env_vars().items():
            cmd = cmd.replace("{{image.env." + str(k) + "}}", str(v if v else ""))

        return cmd

    def get_env_vars(self):
        environment = {}

        for var in self._image["variables"]:
            value = self._database.query(
                f"SELECT value FROM variables WHERE server_id = '{self._server['id']}' AND variable = '{var['variable']}'",
                fetchone=True,
            )["value"]

            if not value:
                continue

            if var["install_only"] and not self._install:
                continue

            environment[var["variable"]] = value

        environment["SERVER_MEMORY"] = self._server["memory"]
        environment["SERVER_PORT"] = self._server["port"]

        return environment



# -*- coding: utf-8 -*-

# Wilfred
# Copyright (C) 2020, Vilhelm Prytz, <vilhelm@prytznet.se>
#
# Licensed under the terms of the MIT license, see LICENSE.
# https://github.com/wilfred-dev/wilfred

import click
import codecs
import locale
import os
import sys

from yaspin import yaspin
from pathlib import Path

from wilfred.docker_conn import docker_client
from wilfred.version import version
from wilfred.config_parser import Config
from wilfred.database import Database
from wilfred.servers import Servers
from wilfred.images import Images
from wilfred.message_handler import warning, error
from wilfred.core import is_integer, random_string

if sys.platform.startswith("win"):
    click.echo("Wilfred does not support Windows")
    sys.exit(1)

config = Config()
database = Database()
images = Images()
servers = Servers(database, docker_client(), config.configuration, images)


def print_version(ctx, param, value):
    """
    print version and exit
    """

    if not value or ctx.resilient_parsing:
        return

    click.echo(
        "✨ wilfred version {}".format(
            "development (0.0.0.dev0)" if version == "0.0.0.dev0" else version
        )
    )
    ctx.exit()


def print_path(ctx, param, value):
    """
    print config/data paths
    """

    if not value or ctx.resilient_parsing:
        return

    click.echo(f"Configuration file: {click.format_filename(config.config_path)}")
    click.echo(f"Image config file: {click.format_filename(images.image_dir)}")
    click.echo(f"Database file: {click.format_filename(database.database_path)}")
    if config.configuration:
        click.echo(
            f"Server data: {click.format_filename(config.configuration['data_path'])}"
        )

    ctx.exit()


def main():
    # snap packages raise some weird ASCII codec errors, so we just force C.UTF-8
    if (
        codecs.lookup(locale.getpreferredencoding()).name == "ascii"
        and os.name == "posix"
    ):
        os.environ["LC_ALL"] = "C.UTF-8"
        os.environ["LANG"] = "C.UTF-8"

    cli()


@click.group()
@click.option(
    "--version", is_flag=True, callback=print_version, expose_value=False, is_eager=True
)
@click.option(
    "--path", is_flag=True, callback=print_path, expose_value=False, is_eager=True
)
def cli():
    """
    Wilfred

    🐿️  A CLI for managing game servers using Docker.

    ⚠️  Wilfred is currently under development and should not be considered stable.
    Features may break or may not be implemented yet. Use with caution.
    """

    pass


@cli.command()
def setup():
    """Setup wilfred, create configuration."""

    if config.configuration:
        warning("A configuration file for Wilfred already exists.")
        click.confirm("Are you sure you wan't to continue?", abort=True)

    data_path = click.prompt(
        "Path for storing server data",
        default=f"{str(Path.home())}/wilfred-data/servers",
    )

    config.write(data_path)


@cli.command("servers")
def servers_list():
    """List all existing servers."""

    click.echo(servers.pretty())


@cli.command("images")
@click.option("--refresh", help="Download the default images from GitHub", is_flag=True)
def list_images(refresh):
    """List images available on file."""

    if refresh:
        with yaspin(text="Refreshing images", color="yellow") as spinner:
            images.download_default(read=True)

            spinner.ok("✅")

    click.echo(images.pretty())


@cli.command()
@click.option(
    "--console",
    help="Attach to server console immediately after creation.",
    is_flag=True,
)
@click.option(
    "--detach", help="Immediately detach during install.", is_flag=True,
)
@click.pass_context
def create(ctx, console, detach):
    """Create a new server."""

    if not config.configuration:
        error("Wilfred has not been configured", exit_code=1)

    name = click.prompt("Server Name").lower()

    if " " in name:
        error("space not allowed in name", exit_code=1)

    click.secho("Available Images", bold=True)
    click.echo(images.pretty())

    image_uid = click.prompt("Image UID", default="minecraft-vanilla")

    if " " in image_uid:
        error("space not allowed in image_uid", exit_code=1)

    if not images.get_image(image_uid):
        error("image does not exist", exit_code=1)

    port = click.prompt("Port", default=25565)
    memory = click.prompt("Memory", default=1024)

    click.secho("Environment Variables", bold=True)

    # create
    database.query(
        " ".join(
            (
                "INSERT INTO servers",
                "(id, name, image_uid, memory, port, custom_startup, status)"
                f"VALUES ('{random_string()}', '{name}', '{image_uid}', '{memory}', '{port}', NULL, 'installing')",
            )
        )
    )

    server = database.query(
        f"SELECT * FROM servers WHERE name = '{name}'", fetchone=True
    )

    # environment variables available for the container
    for v in images.get_image(image_uid)["variables"]:
        if not v["hidden"]:
            value = click.prompt(
                v["prompt"], default=v["default"] if v["default"] is not True else None
            )

        if v["hidden"]:
            value = v["default"]

        database.query(
            f"INSERT INTO variables (server_id, variable, value) VALUES ('{server['id']}', '{v['variable']}', '{value}')"
        )

    # custom startup command
    if click.confirm("Would you like to set a custom startup command (optional)?"):
        custom_startup = click.prompt(
            "Custom startup command", default=images.get_image(image_uid)["command"]
        )

        database.query(
            f"UPDATE servers SET custom_startup = '{custom_startup}' WHERE id='{server['id']}'"
        )

    with yaspin(text="Creating server", color="yellow") as spinner:
        servers.install(server, skip_wait=True if detach else False, spinner=spinner)
        spinner.ok("✅ ")

    if console:
        ctx.invoke(start, name=name)
        ctx.invoke(server_console, name=name)


@cli.command("sync")
def sync_cmd():
    """
    Sync all servers on file with Docker (start/stop/kill/create).
    """

    with yaspin(text="Docker sync", color="yellow") as spinner:
        if not config.configuration:
            spinner.fail("💥 Wilfred has not been configured")

        servers.sync()

        spinner.ok("✅ ")


@cli.command()
@click.argument("name")
@click.option(
    "--console", help="Attach to server console immediately after start.", is_flag=True
)
@click.pass_context
def start(ctx, name, console):
    """
    Start server by specifiying the
    name of the server as argument.
    """

    servers.sync(db_update=True)

    with yaspin(text="Server start", color="yellow") as spinner:
        if not config.configuration:
            spinner.fail("💥 Wilfred has not been configured")
            sys.exit(1)

        server = servers.get_by_name(name.lower())

        if not server:
            spinner.fail("💥 Server does not exit")
            sys.exit(1)

        if server["status"] == "installing":
            spinner.fail("💥 Server is installing, start blocked.")
            sys.exit(1)

        servers.set_status(server, "running")
        servers.sync()

        spinner.ok("✅ ")

        if console:
            ctx.invoke(server_console, name=name)


@cli.command()
@click.argument("name")
def kill(name):
    """
    Forcefully kill running server.
    """

    if click.confirm(
        "Are you sure you want to do this? This will kill the running container without saving data."
    ):
        with yaspin(text="Killing server", color="yellow") as spinner:
            if not config.configuration:
                spinner.fail("💥 Wilfred has not been configured")
                sys.exit(1)

            server = servers.get_by_name(name.lower())

            if not server:
                spinner.fail("💥 Server does not exit")
                sys.exit(1)

            servers.kill(server)
            servers.set_status(server, "stopped")
            servers.sync()

            spinner.ok("✅ ")


@cli.command()
@click.argument("name")
def stop(name):
    """
    Stop server.
    """

    servers.sync()

    with yaspin(text="Stopping server", color="yellow") as spinner:
        if not config.configuration:
            spinner.fail("💥 Wilfred has not been configured")
            sys.exit(1)

        server = servers.get_by_name(name.lower())

        if not server:
            spinner.fail("💥 Server does not exit")
            sys.exit(1)

        if server["status"] == "installing":
            spinner.fail(
                " ".join(
                    (
                        "💥 Server is installing, you cannot gracefully stop it.",
                        "Use `wilfred kill` if the installation process has hanged.",
                    )
                )
            )
            sys.exit(1)

        servers.set_status(server, "stopped")
        servers.sync()

        spinner.ok("✅ ")


@cli.command()
@click.argument("name")
def delete(name):
    """
    Delete existing server.
    """

    if click.confirm(
        "Are you sure you want to do this? All data will be permanently deleted."
    ):
        with yaspin(text="Deleting server", color="yellow") as spinner:
            if not config.configuration:
                spinner.fail("💥 Wilfred has not been configured")
                sys.exit(1)

            server = servers.get_by_name(name.lower())

            if not server:
                spinner.fail("💥 Server does not exit")
                sys.exit(1)

            servers.remove(server)
            spinner.ok("✅ ")


@cli.command("console")
@click.argument("name")
def server_console(name):
    """
    Attach to server console, view log and run commands.
    """

    if not config.configuration:
        error("Wilfred has not been configured", exit_code=1)

    server = servers.get_by_name(name.lower())

    if not server:
        error("Server does not exit", exit_code=1)

    click.secho(
        " ".join(
            (
                f"Viewing server console of {name} (id {server['id']})",
                f"{'- input disabled, installing' if server['status'] == 'installing' else ''}",
            )
        ),
        bold=True,
    )

    servers.console(
        server, disable_user_input=True if server["status"] == "installing" else False
    )


@cli.command()
@click.argument("name")
def edit(name):
    """
    Edit server (name, memory, port, environment variables)
    """

    if not config.configuration:
        error("Wilfred has not been configured", exit_code=1)

    server = servers.get_by_name(name.lower())

    if not server:
        error("Server does not exist", exit_code=1)

    click.echo(servers.pretty(server=server))
    click.echo("Leave values empty to use existing value")

    name = click.prompt("Name", default=server["name"]).lower()

    if " " in name:
        error("space not allowed in name", exit_code=1)

    port = click.prompt("Port", default=server["port"])
    memory = click.prompt("Memory", default=server["memory"])

    if not is_integer(port) or not is_integer(memory):
        error("port/memory must be integer", exit_code=1)

    click.secho("Environment Variables", bold=True)

    for v in images.get_image(server["image_uid"])["variables"]:
        if v["install_only"]:
            continue

        curr = database.query(
            f"SELECT * FROM variables WHERE server_id = '{server['id']}' AND variable = '{v['variable']}'",
            fetchone=True,
        )

        if not curr:
            continue

        if not v["hidden"]:
            value = click.prompt(v["prompt"], default=curr["value"])

        if v["hidden"]:
            value = v["default"]

        database.query(
            " ".join(
                (
                    f"UPDATE variables SET value = '{value}'",
                    f"WHERE server_id = '{server['id']}' AND variable = '{v['variable']}'",
                )
            )
        )

    server["custom_startup"] = (
        None if server["custom_startup"] == "None" else server["custom_startup"]
    )
    custom_startup = None

    if server["custom_startup"] is not None:
        custom_startup = click.prompt(
            "Custom startup command", default=server["custom_startup"]
        )

    if server["custom_startup"] is None:
        if click.confirm("Would you like to set a custom startup command?"):
            custom_startup = click.prompt(
                "Custom startup command",
                default=images.get_image(server["image_uid"])["command"],
            )

    database.query(f"UPDATE servers SET name = '{name}' WHERE id='{server['id']}'")
    database.query(f"UPDATE servers SET port = {port} WHERE id='{server['id']}'")
    database.query(f"UPDATE servers SET memory = {memory} WHERE id='{server['id']}'")

    if custom_startup:
        database.query(
            f"UPDATE servers SET custom_startup = '{custom_startup}' WHERE id='{server['id']}'"
        )

    click.echo(
        "✅ Server information updated, restart server for changes to take effect"
    )


if __name__ == "__main__":
    main()



# Wilfred
# Copyright (C) 2020, Vilhelm Prytz, <vilhelm@prytznet.se>
#
# Licensed under the terms of the MIT license, see LICENSE.
# https://github.com/wilfred-dev/wilfred

import click
import docker

from tabulate import tabulate
from pathlib import Path
from shutil import rmtree
from os import remove as remove_file
from time import sleep

from wilfred.message_handler import error
from wilfred.keyboard import KeyboardThread
from wilfred.container_variables import ContainerVariables


class Servers(object):
    def __init__(self, database, docker_client, configuration, images):
        self._database = database
        self._images = images
        self._configuration = configuration
        self._docker_client = docker_client

        self._get_db_servers()

    def pretty(self, server=None):
        self._running_docker_sync()

        servers = [server] if server else self._servers

        for server in servers:
            server.update(
                (
                    k,
                    str(v)
                    .replace("running", click.style("running", fg="green"))
                    .replace("stopped", click.style("stopped", fg="red"))
                    .replace("installing", click.style("installing", fg="yellow")),
                )
                for k, v in server.items()
            )

        headers = {
            "id": click.style("ID", bold=True),
            "name": click.style("Name", bold=True),
            "image_uid": click.style("Image UID", bold=True),
            "memory": click.style("Memory (RAM)", bold=True),
            "port": click.style("Port", bold=True),
            "status": click.style("Status", bold=True),
            "custom_startup": click.style("Custom startup command", bold=True),
        }

        return tabulate(servers, headers=headers, tablefmt="fancy_grid",)

    def set_status(self, server, status):
        self._database.query(
            f"UPDATE servers SET status = '{status}' WHERE id = '{server['id']}'"
        )
        self._get_db_servers()

    def sync(self, db_update=False):
        if db_update:
            self._get_db_servers()

        with click.progressbar(
            self._servers, label="Syncing servers", length=len(self._servers)
        ) as servers:
            for server in servers:
                if server["status"] == "installing":
                    try:
                        self._docker_client.containers.get(f"wilfred_{server['id']}")
                    except docker.errors.NotFound:
                        self.set_status(server, "stopped")

                # stopped
                if server["status"] == "stopped":
                    self._stop(server)

                # start
                if server["status"] == "running":
                    try:
                        self._docker_client.containers.get(f"wilfred_{server['id']}")
                    except docker.errors.NotFound:
                        self._start(server)

    def get_by_name(self, name):
        self._get_db_servers()

        server = list(filter(lambda x: x["name"] == name, self._servers))

        return server[0] if server else None

    def remove(self, server):
        path = f"{self._configuration['data_path']}/{server['id']}"

        self._database.query(f"DELETE FROM servers WHERE id='{server['id']}'")
        self._database.query(f"DELETE FROM variables WHERE server_id='{server['id']}'")

        try:
            container = self._docker_client.containers.get(f"wilfred_{server['id']}")
            container.stop()
        except docker.errors.NotFound:
            pass

        rmtree(path, ignore_errors=True)

    def console(self, server, disable_user_input=False):
        try:
            container = self._docker_client.containers.get(f"wilfred_{server['id']}")
        except docker.errors.NotFound:
            error("server is not running", exit_code=1)

        if not disable_user_input:
            KeyboardThread(self._console_input_callback, params=server)

        for line in container.logs(stream=True, tail=200):
            click.echo(line.strip())

    def _console_input_callback(self, payload, server):
        self._command(server, payload)

    def _command(self, server, command):
        try:
            container = self._docker_client.containers.get(f"wilfred_{server['id']}")
        except docker.errors.NotFound:
            error("server is not running", exit_code=1)

        try:
            s = container.attach_socket(params={"stdin": 1, "stream": 1})
            s._sock.send(f"{command}\n".encode("utf-8"))
            s.close()
        except Exception as e:
            error(
                f"unable to send command '{command}' on server {server['id']}, err {click.style(str(e), bold=True)}",
                exit_code=1,
            )

    def _get_db_servers(self):
        self._servers = self._database.query("SELECT * FROM servers")

    def _running_docker_sync(self):
        for server in self._servers:
            try:
                self._docker_client.containers.get(f"wilfred_{server['id']}")
            except docker.errors.NotFound:
                self.set_status(server, "stopped")

        self._get_db_servers()

    def _parse_startup_command(self, cmd, server, image):
        return ContainerVariables(server, image, self._database).parse_startup_command(
            cmd.replace("{{SERVER_MEMORY}}", str(server["memory"])).replace(
                "{{SERVER_PORT}}", str(server["port"])
            )
        )

    def install(self, server, skip_wait=False, spinner=None):
        path = f"{self._configuration['data_path']}/{server['id']}"
        image = self._images.get_image(server["image_uid"])

        try:
            Path(path).mkdir(parents=True, exist_ok=True)
        except Exception as e:
            error(
                f"unable to create server data directory, {click.style(str(e), bold=True)}",
                exit_code=1,
            )

        with open(f"{path}/install.sh", "w") as f:
            f.write("cd /server\n" + "\n".join(image["installation"]["script"]))

        if spinner:
            spinner.write(
                "> Pulling Docker image and creating installation container, do not exit"
            )

        try:
            self._docker_client.containers.run(
                image["installation"]["docker_image"],
                f"{image['installation']['shell']} /server/install.sh",
                volumes={path: {"bind": "/server", "mode": "rw"}},
                name=f"wilfred_{server['id']}",
                environment=ContainerVariables(
                    server, image, self._database, install=True
                ).get_env_vars(),
                remove=True,
                detach=True,
            )
        except Exception as e:
            self._database.query(f"DELETE FROM servers WHERE id='{server['id']}'")
            error(
                f"unable to create installation Docker container, server removed {click.style(str(e), bold=True)}",
                exit_code=1,
            )

        if skip_wait and spinner:
            spinner.write(
                "> Installation will continue in background, use `wilfred servers` to see if process has finished."
            )

        if not skip_wait:
            if spinner:
                spinner.write(
                    "> You can safely press CTRL+C, the installation will continue in the background."
                )
                spinner.write(
                    "> Run `wilfred servers` too see when the status changes from `installing` to `stopped`."
                )
                spinner.write(
                    f"> You can also follow the installation log using `wilfred console {server['name']}`"
                )
            while self._container_alive(server):
                sleep(1)

    def _container_alive(self, server):
        try:
            self._docker_client.containers.get(f"wilfred_{server['id']}")
        except docker.errors.NotFound:
            return False

        return True

    def _start(self, server):
        path = f"{self._configuration['data_path']}/{server['id']}"
        image = self._images.get_image(server["image_uid"])

        try:
            remove_file(f"{path}/install.sh")
        except Exception:
            pass

        try:
            self._docker_client.containers.run(
                image["docker_image"],
                self._parse_startup_command(server["custom_startup"], server, image)
                if server["custom_startup"] is not None
                else f"{self._parse_startup_command(image['command'], server, image)}",
                volumes={path: {"bind": "/server", "mode": "rw"}},
                name=f"wilfred_{server['id']}",
                remove=True,
                ports={f"{server['port']}/tcp": server["port"]},
                detach=True,
                working_dir="/server",
                mem_limit=f"{server['memory']}m",
                oom_kill_disable=True,
                stdin_open=True,
                environment=ContainerVariables(
                    server, image, self._database
                ).get_env_vars(),
                user=image["user"] if image["user"] else "root",
            )
        except Exception as e:
            error(
                f"unable to start Docker container {click.style(str(e), bold=True)}",
                exit_code=1,
            )

    def kill(self, server):
        try:
            container = self._docker_client.containers.get(f"wilfred_{server['id']}")
        except docker.errors.NotFound:
            return

        container.kill()

    def _stop(self, server):
        image = self._images.get_image(server["image_uid"])

        try:
            container = self._docker_client.containers.get(f"wilfred_{server['id']}")
        except docker.errors.NotFound:
            return

        if not image["stop_command"]:
            try:
                container.stop()
            except Exception as e:
                exit(
                    f"could not stop container {click.style(str(e), bold=True)}",
                    exit_code=1,
                )

            return

        self._command(server, image["stop_command"])

        stopped = False

        while not stopped:
            try:
                self._docker_client.containers.get(f"wilfred_{server['id']}")
            except docker.errors.NotFound:
                stopped = True
